from marshmallow import fields

from pynamodb import attributes

PYNAMODB_TYPE_MAPPING = {
    attributes.NumberAttribute: fields.Number,
    attributes.JSONAttribute: fields.Raw,
    attributes.UnicodeAttribute: fields.String,
    attributes.BooleanAttribute: fields.Boolean,
    attributes.UTCDateTimeAttribute: fields.DateTime,
    attributes.MapAttributeMeta: {True: fields.Nested,
                                  False: fields.Dict},
    attributes.ListAttribute: {True: fields.List,
                               False: fields.Raw}
}


def attribute2field(attribute, validate_collections):
    try:
        pynamodb_type = type(attribute)
        field = PYNAMODB_TYPE_MAPPING[pynamodb_type]
    except KeyError:
        pynamodb_type = type(pynamodb_type)
        field = PYNAMODB_TYPE_MAPPING[pynamodb_type]

    if isinstance(attribute, attributes.ListAttribute) or isinstance(attribute, attributes.MapAttribute):
        return field[validate_collections]
    else:
        return field



from marshmallow import Schema, SchemaOpts, post_load, fields
from marshmallow.schema import SchemaMeta
from marshmallow_pynamodb.convert import attribute2field
from pynamodb.attributes import Attribute
from six import with_metaclass


class ModelOpts(SchemaOpts):
    def __init__(self, meta):
        SchemaOpts.__init__(self, meta)
        self.model = getattr(meta, 'model', None)
        self.validate = getattr(meta, 'validate', False)


class ModelMeta(SchemaMeta):
    @classmethod
    def get_declared_fields(mcs, klass, cls_fields, inherited_fields, dict_cls):
        declared_fields = super(ModelMeta, mcs).get_declared_fields(klass, cls_fields, inherited_fields, dict_cls)
        if klass.opts.model:
            attributes = {name: attr for name, attr in vars(klass.opts.model).iteritems() if
                          isinstance(attr, Attribute)}
            klass.opts.model.attributes = dict()
            for attr_name, attribute in attributes.iteritems():
                field = attribute2field(attribute, klass.opts.validate)

                if field == fields.Nested:
                    instance_of = type(attribute)

                    class Meta:
                        model = instance_of
                        validate = True
                    sub_model = type(instance_of.__name__, (ModelSchema, ), {'Meta': Meta})
                    field = field(sub_model)
                elif field == fields.List:
                    class Meta:
                        model = attribute.element_type
                        validate = True
                    element_type = type(attribute.element_type.__name__, (ModelSchema, ), {'Meta': Meta})
                    field = field(fields.Nested(element_type))
                else:
                    field = field()

                field_name = attribute.attr_name if attribute.attr_name else attr_name
                if attribute.is_hash_key or attribute.is_range_key or not attribute.null:
                    field.required = True
                    if attribute.is_hash_key:
                        klass.opts.hash_key = field_name
                    elif attribute.is_range_key:
                        klass.opts.range_key = field_name

                declared_fields[field_name] = field
        return declared_fields


class ModelSchema(with_metaclass(ModelMeta, Schema)):
    OPTIONS_CLASS = ModelOpts

    @post_load(pass_original=True)
    def hydrate_pynamo_model(self, data, orig_data):
        hash_key = orig_data.pop(getattr(self.opts, 'hash_key', None), None)
        range_key = orig_data.pop(getattr(self.opts, 'range_key', None), None)
        return self.opts.model(hash_key=hash_key, range_key=range_key, **orig_data)



from marshmallow_pynamodb.schema import ModelSchema

from pynamodb.attributes import ListAttribute, MapAttribute, NumberAttribute, UnicodeAttribute
from pynamodb.models import Model

from unittest import TestCase


class Location(MapAttribute):
    lat = NumberAttribute(attr_name='latitude')
    lng = NumberAttribute(attr_name='longitude')
    name = UnicodeAttribute()


class Person(MapAttribute):
    fname = UnicodeAttribute(attr_name='firstName')
    lname = UnicodeAttribute(attr_name='lastName')
    age = NumberAttribute()


class OfficeEmployeeMap(MapAttribute):
    office_employee_id = NumberAttribute()
    person = Person()
    office_location = Location()


class Office(Model):
    class Meta:
        table_name = 'OfficeModel'
        host = "http://localhost:8000"

    office_id = NumberAttribute(hash_key=True)
    address = Location()
    employees = ListAttribute(of=OfficeEmployeeMap)


class OfficeSchema(ModelSchema):
    class Meta:
        model = Office
        validate = True


hash_key = 789
attrs = {'address': {'latitude': 6.98454,
                     'longitude': 172.38832,
                     'name': 'some_location'},
         'employees': [{'office_employee_id': 123,
                        'person': {'firstName': 'John',
                                   'lastName': 'Smith',
                                   'age': 45},
                        'office_location': {'latitude': -24.0853,
                                            'longitude': 144.87660,
                                            'name': 'other_location'}},
                       {'office_employee_id': 456,
                        'person': {'firstName': 'Jane',
                                   'lastName': 'Doe',
                                   'age': 33},
                        'office_location': {'latitude': -20.57989,
                                            'longitude': 92.30463,
                                            'name': 'yal'}}]}


class TestValidatedModelSchema(TestCase):
    def setUp(self):
        if not Office.exists():
            Office.create_table(read_capacity_units=1, write_capacity_units=1)

    def test_load(self):
        expected_model = Office(hash_key=hash_key, **attrs)
        attrs['office_id'] = hash_key

        data, errors = OfficeSchema().load(attrs)

        self.assertTrue(getattr(OfficeSchema, '_declared_fields')['office_id'].required)
        self.assertDictEqual(errors, {})
        self.assertDictEqual(data.attribute_values, expected_model.attribute_values)

    def test_load_fails_dict_level_0(self):
        attrs['office_id'] = 'foobar'

        data, errors = OfficeSchema().load(attrs)

        self.assertDictEqual(errors, {'office_id': [u'Not a valid number.']})

    def test_load_fails_dict_level_gt_0(self):
        attrs['office_id'] = hash_key
        attrs['employees'][0]['office_location']['latitude'] = 'foobar'

        data, errors = OfficeSchema().load(attrs)

        self.assertDictEqual(errors, {'employees': {0: {'office_location': {'latitude': [u'Not a valid number.']}}}})

    def test_dump(self):
        model = Office(hash_key=hash_key, **attrs)
        attrs['office_id'] = hash_key

        data, errors = OfficeSchema().dump(model)

        self.assertDictEqual(errors, {})
        self.assertDictEqual(data, attrs)



from setuptools import setup, find_packages

setup(
    name="marshmallow-pynamodb",
    version="0.2.7",
    packages=find_packages(),
    description='PynamoDB integration with the marshmallow (de)serialization library',
    author='Mathew Marcus',
    author_email='mathewmarcus456@gmail.com',
    long_description=open('README.rst').read(),
    install_requires=[
        "marshmallow>=2.12.2",
        "pynamodb>=2.0.3",
    ]
)
